=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2006-12-10 18:46:53 +0000
+++ bzrlib/errors.py	2006-12-11 03:11:05 +0000
@@ -1611,3 +1611,15 @@
 
     def __init__(self, other_tree):
         self.other_tree = other_tree
+
+
+class BadReferenceTarget(BzrError):
+
+    _fmt = "Can't add reference to %(other_tree)s into %(tree)s.  %(reason)s"
+
+    internal_error = True
+
+    def __init__(self, tree, other_tree, reason):
+        self.tree = tree
+        self.other_tree = other_tree
+        self.reason = reason

=== modified file 'bzrlib/tests/workingtree_implementations/__init__.py'
--- bzrlib/tests/workingtree_implementations/__init__.py	2006-12-10 07:39:19 +0000
+++ bzrlib/tests/workingtree_implementations/__init__.py	2006-12-11 03:11:05 +0000
@@ -51,6 +51,7 @@
 def test_suite():
     result = TestSuite()
     test_workingtree_implementations = [
+        'bzrlib.tests.workingtree_implementations.test_add_reference',
         'bzrlib.tests.workingtree_implementations.test_basis_inventory',
         'bzrlib.tests.workingtree_implementations.test_break_lock',
         'bzrlib.tests.workingtree_implementations.test_changes_from',

=== added file 'bzrlib/tests/workingtree_implementations/test_add_reference.py'
--- bzrlib/tests/workingtree_implementations/test_add_reference.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/workingtree_implementations/test_add_reference.py	2006-12-11 03:11:05 +0000
@@ -0,0 +1,83 @@
+# Copyright (C) 2006 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+import os
+
+from bzrlib import errors, tests, workingtree
+from bzrlib.tests.workingtree_implementations import TestCaseWithWorkingTree
+
+class TestBasisInventory(TestCaseWithWorkingTree):
+
+    def make_trees(self):
+        tree = self.make_branch_and_tree('tree')
+        tree.set_root_id('root-id')
+        self.build_tree(['tree/file1'])
+        tree.add('file1', 'file1-id')
+        sub_tree = self.make_branch_and_tree('tree/sub-tree')
+        sub_tree.set_root_id('sub-tree-root-id')
+        return tree, sub_tree
+
+    def test_add_reference(self):
+        tree, sub_tree = self.make_trees()
+        try:
+            tree.add_reference(sub_tree)
+        except errors.UnsupportedOperation:
+            assert tree.__class__ in (workingtree.WorkingTree2, 
+                                      workingtree.WorkingTree3)
+            raise tests.TestSkipped('Tree format does not support references')
+        self.assertEqual(tree.path2id('sub-tree'), 'sub-tree-root-id')
+        self.assertEqual(tree.inventory['sub-tree-root-id'].kind, 
+                         'tree-reference')
+
+    def test_add_reference_same_root(self):
+        tree = self.make_branch_and_tree('tree')
+        self.build_tree(['tree/file1'])
+        tree.add('file1', 'file1-id')
+        tree.set_root_id('root-id')
+        sub_tree = self.make_branch_and_tree('tree/sub-tree')
+        sub_tree.set_root_id('root-id')
+        try:
+            self.assertRaises(errors.BadReferenceTarget, tree.add_reference, 
+                              sub_tree)
+        except errors.UnsupportedOperation:
+            assert tree.__class__ in (workingtree.WorkingTree2, 
+                                      workingtree.WorkingTree3)
+            raise tests.TestSkipped('Tree format does not support references')
+
+    def test_root_present(self):
+        """Subtree root is present, though not the working tree root"""
+        tree, sub_tree = self.make_trees()
+        sub_tree.set_root_id('file1-id')
+        try:
+            self.assertRaises(errors.BadReferenceTarget, tree.add_reference, 
+                              sub_tree)
+        except errors.UnsupportedOperation:
+            assert tree.__class__ in (workingtree.WorkingTree2, 
+                                      workingtree.WorkingTree3)
+            raise tests.TestSkipped('Tree format does not support references')
+
+    def test_add_non_subtree(self):
+        tree, sub_tree = self.make_trees()
+        os.rename('tree/sub-tree', 'sibling')
+        sibling = workingtree.WorkingTree.open('sibling')
+        try:
+            self.assertRaises(errors.BadReferenceTarget, tree.add_reference, 
+                              sibling)
+        except errors.UnsupportedOperation:
+            assert tree.__class__ in (workingtree.WorkingTree2, 
+                                      workingtree.WorkingTree3)
+            raise tests.TestSkipped('Tree format does not support references')
+

=== modified file 'bzrlib/workingtree.py'
--- bzrlib/workingtree.py	2006-12-10 18:46:53 +0000
+++ bzrlib/workingtree.py	2006-12-11 03:11:05 +0000
@@ -84,7 +84,7 @@
                            MergeModifiedFormatError,
                            UnsupportedOperation,
                            )
-from bzrlib.inventory import InventoryEntry, Inventory, ROOT_ID
+from bzrlib.inventory import InventoryEntry, Inventory, ROOT_ID, TreeReference
 from bzrlib.lockable_files import LockableFiles, TransportLock
 from bzrlib.lockdir import LockDir
 import bzrlib.mutabletree
@@ -593,6 +593,10 @@
                 inv.add_path(f, kind=kind, file_id=file_id)
         self._write_inventory(inv)
 
+    def add_reference(self, sub_tree):
+        """Add a TreeReference to the tree, pointing at sub_tree"""
+        raise UnsupportedOperation(self.add_reference, self)
+
     @needs_tree_write_lock
     def _gather_kinds(self, files, kinds):
         """See MutableTree._gather_kinds."""
@@ -1991,7 +1995,25 @@
 
 
 class WorkingTree4(WorkingTree3):
-    pass
+
+    def add_reference(self, sub_tree):
+        try:
+            sub_tree_path = self.relpath(sub_tree.basedir)
+        except errors.PathNotChild:
+            raise errors.BadReferenceTarget(self, sub_tree, 
+                                            'Target not inside tree.')
+        parent_id = self.path2id(osutils.dirname(sub_tree_path))
+        name = osutils.basename(sub_tree_path)
+        sub_tree_id = sub_tree.get_root_id()
+        if sub_tree_id == self.get_root_id():
+            raise errors.BadReferenceTarget(self, sub_tree, 
+                                     'Trees have the same root id.')
+        if sub_tree_id in self.inventory:
+            raise errors.BadReferenceTarget(self, sub_tree, 
+                                            'Root id already present in tree')
+        entry = TreeReference(sub_tree_id, name, parent_id, None, 
+                              None)
+        self.inventory.add(entry)
 
 
 def get_conflicted_stem(path):
@@ -2213,7 +2235,7 @@
         if revision_id is None:
             revision_id = branch.last_revision()
         inv = Inventory(root_id=generate_ids.gen_root_id())
-        wt = WorkingTree3(a_bzrdir.root_transport.local_abspath('.'),
+        wt = self._tree_class(a_bzrdir.root_transport.local_abspath('.'),
                          branch,
                          inv,
                          _internal=True,

